@@ -20,7 +20,7 @@ DEFAULT_KINDS = [
     {"role": "**/roles/*/"},
     {"tasks": "**/tasks/*.{yaml,yml}"},
     {"handlers": "**/handlers/*.{yaml,yml}"},
-    {"vars": "**/{vars,defaults}/*.{yaml,yml}"},
+    {"vars": "**/{host_vars,group_vars,vars,defaults}/**/*.{yaml,yml}"},
     {"meta": "**/meta/main.{yaml,yml}"},
     {"yaml": ".config/molecule/config.{yaml,yml}"},  # molecule global config
     {"yaml": "**/molecule/*/molecule.{yaml,yml}"},  # molecule config
@@ -50,6 +50,7 @@ options = Namespace(
     mock_roles=[],
     loop_var_prefix=None,
     offline=False,
+    extra_vars=None,
 )
 
 # Used to store detected tag deprecations
@@ -64,7 +65,7 @@ def ansible_collections_path() -> str:
     """Return collection path variable for current version of Ansible."""
     # respect Ansible behavior, which is to load old name if present
     for env_var in ["ANSIBLE_COLLECTIONS_PATHS", "ANSIBLE_COLLECTIONS_PATH"]:
-        if "ANSIBLE_COLLECTIONS_PATHS" in os.environ:
+        if env_var in os.environ:
             return env_var
 
     # https://github.com/ansible/ansible/pull/70007
