@@ -23,13 +23,11 @@ import contextlib
 import inspect
 import logging
 import os
-import pprint
-import subprocess
 from argparse import Namespace
-from collections import OrderedDict
+from collections.abc import ItemsView
 from functools import lru_cache
 from pathlib import Path
-from typing import Callable, ItemsView, List, Optional, Tuple
+from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Union
 
 import yaml
 from ansible import constants
@@ -39,17 +37,32 @@ from ansible.parsing.mod_args import ModuleArgsParser
 from ansible.parsing.splitter import split_args
 from ansible.parsing.yaml.constructor import AnsibleConstructor
 from ansible.parsing.yaml.loader import AnsibleLoader
-from ansible.parsing.yaml.objects import AnsibleSequence
+from ansible.parsing.yaml.objects import AnsibleBaseYAMLObject, AnsibleSequence
 from ansible.plugins.loader import add_all_plugin_dirs
 from ansible.template import Templar
+
+try:
+    from ansible.module_utils.parsing.convert_bool import boolean
+except ImportError:
+    try:
+        from ansible.utils.boolean import boolean
+    except ImportError:
+        try:
+            from ansible.utils import boolean
+        except ImportError:
+            boolean = constants.mk_boolean
+
 from yaml.composer import Composer
 from yaml.representer import RepresenterError
 
-from ansiblelint.constants import (
-    ANSIBLE_FAILURE_RC, CUSTOM_RULESDIR_ENVVAR, DEFAULT_RULESDIR, FileType,
+from ansiblelint._internal.rules import (
+    AnsibleParserErrorRule,
+    LoadingFailureRule,
+    RuntimeErrorRule,
 )
+from ansiblelint.constants import FileType
 from ansiblelint.errors import MatchError
-from ansiblelint.file_utils import normpath
+from ansiblelint.file_utils import Lintable, get_yaml_files
 
 # ansible-lint doesn't need/want to know about encrypted secrets, so we pass a
 # string as the password to enable such yaml files to be opened and parsed
@@ -75,7 +88,7 @@ def path_dwim(basedir: str, given: str) -> str:
     return dl.path_dwim(given)
 
 
-def ansible_template(basedir, varname, templatevars, **kwargs):
+def ansible_template(basedir: str, varname: Any, templatevars, **kwargs) -> Any:
     dl = DataLoader()
     dl.set_basedir(basedir)
     templar = Templar(dl, variables=templatevars)
@@ -86,14 +99,47 @@ LINE_NUMBER_KEY = '__line__'
 FILENAME_KEY = '__file__'
 
 VALID_KEYS = [
-    'name', 'action', 'when', 'async', 'poll', 'notify',
-    'first_available_file', 'include', 'include_tasks', 'import_tasks', 'import_playbook',
-    'tags', 'register', 'ignore_errors', 'delegate_to',
-    'local_action', 'transport', 'remote_user', 'sudo',
-    'sudo_user', 'sudo_pass', 'when', 'connection', 'environment', 'args', 'always_run',
-    'any_errors_fatal', 'changed_when', 'failed_when', 'check_mode', 'delay',
-    'retries', 'until', 'su', 'su_user', 'su_pass', 'no_log', 'run_once',
-    'become', 'become_user', 'become_method', FILENAME_KEY,
+    'name',
+    'action',
+    'when',
+    'async',
+    'poll',
+    'notify',
+    'first_available_file',
+    'include',
+    'include_tasks',
+    'import_tasks',
+    'import_playbook',
+    'tags',
+    'register',
+    'ignore_errors',
+    'delegate_to',
+    'local_action',
+    'transport',
+    'remote_user',
+    'sudo',
+    'sudo_user',
+    'sudo_pass',
+    'when',
+    'connection',
+    'environment',
+    'args',
+    'any_errors_fatal',
+    'changed_when',
+    'failed_when',
+    'check_mode',
+    'delay',
+    'retries',
+    'until',
+    'su',
+    'su_user',
+    'su_pass',
+    'no_log',
+    'run_once',
+    'become',
+    'become_user',
+    'become_method',
+    FILENAME_KEY,
 ]
 
 BLOCK_NAME_TO_ACTION_TYPE_MAP = {
@@ -107,7 +153,7 @@ BLOCK_NAME_TO_ACTION_TYPE_MAP = {
 }
 
 
-def tokenize(line):
+def tokenize(line: str) -> Tuple[str, List[str], Dict]:
     tokens = line.lstrip().split(" ")
     if tokens[0] == '-':
         tokens = tokens[1:]
@@ -131,20 +177,12 @@ def tokenize(line):
 def _playbook_items(pb_data: dict) -> ItemsView:
     if isinstance(pb_data, dict):
         return pb_data.items()
-    elif not pb_data:
+    if not pb_data:
         return []
-    else:
-        return [item for play in pb_data for item in play.items()]
-
-
-def _rebind_match_filename(filename: str, func) -> Callable:
-    def func_wrapper(*args, **kwargs):
-        try:
-            return func(*args, **kwargs)
-        except MatchError as e:
-            e.filename = filename
-            raise e
-    return func_wrapper
+    # "if play" prevents failure if the play sequence contains None,
+    # which is weird but currently allowed by Ansible
+    # https://github.com/ansible-community/ansible-lint/issues/849
+    return [item for play in pb_data if play for item in play.items()]
 
 
 def _set_collections_basedir(basedir: str):
@@ -163,43 +201,59 @@ def _set_collections_basedir(basedir: str):
         set_collection_playbook_paths(basedir)
 
 
-def find_children(playbook: Tuple[str, str], playbook_dir: str) -> List:
-    if not os.path.exists(playbook[0]):
+def find_children(lintable: Lintable) -> List[Lintable]:  # noqa: C901
+    if not lintable.path.exists():
         return []
-    _set_collections_basedir(playbook_dir or '.')
+    playbook_dir = str(lintable.path.parent)
+    _set_collections_basedir(playbook_dir or os.path.abspath('.'))
     add_all_plugin_dirs(playbook_dir or '.')
-    if playbook[1] == 'role':
-        playbook_ds = {'roles': [{'role': playbook[0]}]}
+    if lintable.kind == 'role':
+        playbook_ds = {'roles': [{'role': str(lintable.path)}]}
+    elif lintable.kind not in ("playbook", "tasks"):
+        return []
     else:
         try:
-            playbook_ds = parse_yaml_from_file(playbook[0])
+            playbook_ds = parse_yaml_from_file(str(lintable.path))
         except AnsibleError as e:
             raise SystemExit(str(e))
     results = []
-    basedir = os.path.dirname(playbook[0])
-    items = _playbook_items(playbook_ds)
-    for item in items:
-        for child in _rebind_match_filename(playbook[0], play_children)(
-                basedir, item, playbook[1], playbook_dir):
-            if "$" in child['path'] or "{{" in child['path']:
+    basedir = os.path.dirname(str(lintable.path))
+    # playbook_ds can be an AnsibleUnicode string, which we consider invalid
+    if isinstance(playbook_ds, str):
+        raise MatchError(filename=str(lintable.path), rule=LoadingFailureRule)
+    for item in _playbook_items(playbook_ds):
+        for child in play_children(basedir, item, lintable.kind, playbook_dir):
+            # We avoid processing parametrized children
+            path_str = str(child.path)
+            if "$" in path_str or "{{" in path_str:
                 continue
+
+            # Repair incorrect paths obtained when old syntax was used, like:
+            # - include: simpletask.yml tags=nginx
             valid_tokens = list()
-            for token in split_args(child['path']):
+            for token in split_args(path_str):
                 if '=' in token:
                     break
                 valid_tokens.append(token)
             path = ' '.join(valid_tokens)
-            results.append({
-                'path': path_dwim(basedir, path),
-                'type': child['type']
-            })
+            if path != path_str:
+                child.path = Path(path)
+                child.name = child.path.name
+
+            results.append(child)
     return results
 
 
-def template(basedir, value, vars, fail_on_undefined=False, **kwargs):
+def template(
+    basedir: str, value: Any, variables, fail_on_undefined=False, **kwargs
+) -> Any:
     try:
-        value = ansible_template(os.path.abspath(basedir), value, vars,
-                                 **dict(kwargs, fail_on_undefined=fail_on_undefined))
+        value = ansible_template(
+            os.path.abspath(basedir),
+            value,
+            variables,
+            **dict(kwargs, fail_on_undefined=fail_on_undefined),
+        )
         # Hack to skip the following exception when using to_json filter on a variable.
         # I guess the filter doesn't like empty vars...
     except (AnsibleError, ValueError, RepresenterError):
@@ -208,8 +262,10 @@ def template(basedir, value, vars, fail_on_undefined=False, **kwargs):
     return value
 
 
-def play_children(basedir, item, parent_type, playbook_dir):
-    delegate_map = {
+def play_children(
+    basedir: str, item: Tuple[str, Any], parent_type, playbook_dir
+) -> List[Lintable]:
+    delegate_map: Dict[str, Callable[[str, Any, Any, FileType], List[Lintable]]] = {
         'tasks': _taskshandlers_children,
         'pre_tasks': _taskshandlers_children,
         'post_tasks': _taskshandlers_children,
@@ -227,15 +283,17 @@ def play_children(basedir, item, parent_type, playbook_dir):
 
     if k in delegate_map:
         if v:
-            v = template(os.path.abspath(basedir),
-                         v,
-                         dict(playbook_dir=PLAYBOOK_DIR or os.path.abspath(basedir)),
-                         fail_on_undefined=False)
+            v = template(
+                os.path.abspath(basedir),
+                v,
+                dict(playbook_dir=PLAYBOOK_DIR or os.path.abspath(basedir)),
+                fail_on_undefined=False,
+            )
             return delegate_map[k](basedir, k, v, parent_type)
     return []
 
 
-def _include_children(basedir, k, v, parent_type):
+def _include_children(basedir: str, k, v, parent_type) -> List[Lintable]:
     # handle special case include_tasks: name=filename.yml
     if k == 'include_tasks' and isinstance(v, dict) and 'file' in v:
         v = v['file']
@@ -246,11 +304,16 @@ def _include_children(basedir, k, v, parent_type):
     result = path_dwim(basedir, args[0])
     if not os.path.exists(result):
         result = path_dwim(os.path.join(os.path.dirname(basedir)), v)
-    return [{'path': result, 'type': parent_type}]
+    return [Lintable(result, kind=parent_type)]
 
 
-def _taskshandlers_children(basedir, k, v, parent_type: FileType) -> List:
-    results = []
+def _taskshandlers_children(basedir, k, v, parent_type: FileType) -> List[Lintable]:
+    results: List[Lintable] = []
+    if v is None:
+        raise MatchError(
+            message="A malformed block was encountered while loading a block.",
+            rule=RuntimeErrorRule(),
+        )
     for th in v:
 
         # ignore empty tasks, `-`
@@ -259,17 +322,28 @@ def _taskshandlers_children(basedir, k, v, parent_type: FileType) -> List:
 
         with contextlib.suppress(LookupError):
             children = _get_task_handler_children_for_tasks_or_playbooks(
-                th, basedir, k, parent_type,
+                th,
+                basedir,
+                k,
+                parent_type,
             )
             results.append(children)
             continue
 
-        if 'include_role' in th or 'import_role' in th:  # lgtm [py/unreachable-statement]
+        if (
+            'include_role' in th or 'import_role' in th
+        ):  # lgtm [py/unreachable-statement]
             th = normalize_task_v2(th)
             _validate_task_handler_action_for_role(th['action'])
-            results.extend(_roles_children(basedir, k, [th['action'].get("name")],
-                                           parent_type,
-                                           main=th['action'].get('tasks_from', 'main')))
+            results.extend(
+                _roles_children(
+                    basedir,
+                    k,
+                    [th['action'].get("name")],
+                    parent_type,
+                    main=th['action'].get('tasks_from', 'main'),
+                )
+            )
             continue
 
         if 'block' not in th:
@@ -277,16 +351,23 @@ def _taskshandlers_children(basedir, k, v, parent_type: FileType) -> List:
 
         results.extend(_taskshandlers_children(basedir, k, th['block'], parent_type))
         if 'rescue' in th:
-            results.extend(_taskshandlers_children(basedir, k, th['rescue'], parent_type))
+            results.extend(
+                _taskshandlers_children(basedir, k, th['rescue'], parent_type)
+            )
         if 'always' in th:
-            results.extend(_taskshandlers_children(basedir, k, th['always'], parent_type))
+            results.extend(
+                _taskshandlers_children(basedir, k, th['always'], parent_type)
+            )
 
     return results
 
 
 def _get_task_handler_children_for_tasks_or_playbooks(
-        task_handler, basedir: str, k, parent_type: FileType,
-) -> dict:
+    task_handler,
+    basedir: str,
+    k,
+    parent_type: FileType,
+) -> Lintable:
     """Try to get children of taskhandler for include/import tasks/playbooks."""
     child_type = k if parent_type == 'playbook' else parent_type
 
@@ -299,10 +380,9 @@ def _get_task_handler_children_for_tasks_or_playbooks(
             if not task_handler:
                 continue
 
-            return {
-                'path': path_dwim(basedir, task_handler[task_handler_key]),
-                'type': child_type,
-            }
+            return Lintable(
+                path_dwim(basedir, task_handler[task_handler_key]), kind=child_type
+            )
 
     raise LookupError(
         f'The node contains none of: {", ".join(task_include_keys)}',
@@ -314,26 +394,32 @@ def _validate_task_handler_action_for_role(th_action: dict) -> None:
     module = th_action['__ansible_module__']
 
     if 'name' not in th_action:
-        raise MatchError(f"Failed to find required 'name' key in {module!s}")
+        raise MatchError(message=f"Failed to find required 'name' key in {module!s}")
 
     if not isinstance(th_action['name'], str):
-        raise RuntimeError(
-            f"Value assigned to 'name' key on '{module!s}' is not a string.",
+        raise MatchError(
+            message=f"Value assigned to 'name' key on '{module!s}' is not a string.",
         )
 
 
-def _roles_children(basedir: str, k, v, parent_type: FileType, main='main') -> list:
-    results = []
+def _roles_children(
+    basedir: str, k, v, parent_type: FileType, main='main'
+) -> List[Lintable]:
+    results: List[Lintable] = []
     for role in v:
         if isinstance(role, dict):
             if 'role' in role or 'name' in role:
                 if 'tags' not in role or 'skip_ansible_lint' not in role['tags']:
-                    results.extend(_look_for_role_files(basedir,
-                                                        role.get('role', role.get('name')),
-                                                        main=main))
+                    results.extend(
+                        _look_for_role_files(
+                            basedir, role.get('role', role.get('name')), main=main
+                        )
+                    )
             elif k != 'dependencies':
-                raise SystemExit('role dict {0} does not contain a "role" '
-                                 'or "name" key'.format(role))
+                raise SystemExit(
+                    'role dict {0} does not contain a "role" '
+                    'or "name" key'.format(role)
+                )
         else:
             results.extend(_look_for_role_files(basedir, role, main=main))
     return results
@@ -347,10 +433,10 @@ def _rolepath(basedir: str, role: str) -> Optional[str]:
         path_dwim(basedir, os.path.join('roles', role)),
         path_dwim(basedir, role),
         # if included from roles/[role]/meta/main.yml
-        path_dwim(
-            basedir, os.path.join('..', '..', '..', 'roles', role)
-        ),
+        path_dwim(basedir, os.path.join('..', '..', '..', 'roles', role)),
         path_dwim(basedir, os.path.join('..', '..', role)),
+        # if checking a role in the current directory
+        path_dwim(basedir, os.path.join('..', role)),
     ]
 
     if constants.DEFAULT_ROLES_PATH:
@@ -374,21 +460,22 @@ def _rolepath(basedir: str, role: str) -> Optional[str]:
     return role_path
 
 
-def _look_for_role_files(basedir: str, role: str, main='main') -> list:
+def _look_for_role_files(basedir: str, role: str, main='main') -> List[Lintable]:
     role_path = _rolepath(basedir, role)
     if not role_path:
         return []
 
     results = []
 
-    for th in ['tasks', 'handlers', 'meta']:
-        current_path = os.path.join(role_path, th)
-        for dir, subdirs, files in os.walk(current_path):
+    for kind in ['tasks', 'meta', 'handlers']:
+        current_path = os.path.join(role_path, kind)
+        for folder, subdirs, files in os.walk(current_path):
             for file in files:
                 file_ignorecase = file.lower()
                 if file_ignorecase.endswith(('.yml', '.yaml')):
-                    thpath = os.path.join(dir, file)
-                    results.append({'path': thpath, 'type': th})
+                    thpath = os.path.join(folder, file)
+                    # TODO(ssbarnea): Find correct way to pass kind: FileType
+                    results.append(Lintable(thpath, kind=kind))  # type: ignore
 
     return results
 
@@ -397,8 +484,8 @@ def rolename(filepath):
     idx = filepath.find('roles/')
     if idx < 0:
         return ''
-    role = filepath[idx + 6:]
-    role = role[:role.find('/')]
+    role = filepath[idx + 6 :]
+    role = role[: role.find('/')]
     return role
 
 
@@ -422,36 +509,21 @@ def _sanitize_task(task: dict) -> dict:
     return result
 
 
-# FIXME: drop noqa once this function is made simpler
-# Ref: https://github.com/ansible/ansible-lint/issues/744
-def normalize_task_v2(task: dict) -> dict:  # noqa: C901
+def normalize_task_v2(task: Dict[str, Any]) -> Dict[str, Any]:
     """Ensure tasks have an action key and strings are converted to python objects."""
     result = dict()
-    if 'always_run' in task:
-        # FIXME(ssbarnea): Delayed import to avoid circular import
-        # See https://github.com/ansible/ansible-lint/issues/880
-        # noqa: # pylint:disable=cyclic-import,import-outside-toplevel
-        from ansiblelint.rules.AlwaysRunRule import AlwaysRunRule
-
-        raise MatchError(
-            rule=AlwaysRunRule,
-            filename=task[FILENAME_KEY],
-            linenumber=task[LINE_NUMBER_KEY])
 
     sanitized_task = _sanitize_task(task)
     mod_arg_parser = ModuleArgsParser(sanitized_task)
     try:
         action, arguments, result['delegate_to'] = mod_arg_parser.parse()
     except AnsibleParserError as e:
-        try:
-            task_info = "%s:%s" % (task[FILENAME_KEY], task[LINE_NUMBER_KEY])
-        except KeyError:
-            task_info = "Unknown"
-        pp = pprint.PrettyPrinter(indent=2)
-        task_pprint = pp.pformat(sanitized_task)
-
-        _logger.critical("Couldn't parse task at %s (%s)\n%s", task_info, e.message, task_pprint)
-        raise SystemExit(ANSIBLE_FAILURE_RC)
+        raise MatchError(
+            rule=AnsibleParserErrorRule(),
+            message=e.message,
+            filename=task.get(FILENAME_KEY, "Unknown"),
+            linenumber=task.get(LINE_NUMBER_KEY, 0),
+        )
 
     # denormalize shell -> command conversion
     if '_uses_shell' in arguments:
@@ -463,8 +535,7 @@ def normalize_task_v2(task: dict) -> dict:  # noqa: C901
             # we don't want to re-assign these values, which were
             # determined by the ModuleArgsParser() above
             continue
-        else:
-            result[k] = v
+        result[k] = v
 
     result['action'] = dict(__ansible_module__=action)
 
@@ -482,13 +553,11 @@ def normalize_task_v2(task: dict) -> dict:  # noqa: C901
     return result
 
 
-# FIXME: drop noqa once this function is made simpler
-# Ref: https://github.com/ansible/ansible-lint/issues/744
 def normalize_task_v1(task):  # noqa: C901
     result = dict()
     for (k, v) in task.items():
         if k in VALID_KEYS or k.startswith('with_'):
-            if k == 'local_action' or k == 'action':
+            if k in ('local_action', 'action'):
                 if not isinstance(v, dict):
                     v = _kv_to_dict(v)
                 v['__ansible_arguments__'] = v.get('__ansible_arguments__', list())
@@ -508,15 +577,15 @@ def normalize_task_v1(task):  # noqa: C901
                         # Keep the line number stored
                         result[k] = v
                         continue
-
-                    else:
-                        # Tasks that include playbooks (rather than task files)
-                        # can get here
-                        # https://github.com/ansible/ansible-lint/issues/138
-                        raise RuntimeError("Was not expecting value %s of type %s for key %s\n"
-                                           "Task: %s. Check the syntax of your playbook using "
-                                           "ansible-playbook --syntax-check" %
-                                           (str(v), type(v), k, str(task)))
+                    # Tasks that include playbooks (rather than task files)
+                    # can get here
+                    # https://github.com/ansible-community/ansible-lint/issues/138
+                    raise RuntimeError(
+                        "Was not expecting value %s of type %s for key %s\n"
+                        "Task: %s. Check the syntax of your playbook using "
+                        "ansible-playbook --syntax-check"
+                        % (str(v), type(v), k, str(task))
+                    )
             v['__ansible_arguments__'] = v.get('__ansible_arguments__', list())
             result['action'] = v
     if 'module' in result['action']:
@@ -532,7 +601,7 @@ def normalize_task_v1(task):  # noqa: C901
     return result
 
 
-def normalize_task(task, filename):
+def normalize_task(task: Dict[str, Any], filename: str) -> Dict[str, Any]:
     ansible_action_type = task.get('__ansible_action_type__', 'task')
     if '__ansible_action_type__' in task:
         del task['__ansible_action_type__']
@@ -542,18 +611,32 @@ def normalize_task(task, filename):
     return task
 
 
-def task_to_str(task):
+def task_to_str(task: Dict[str, Any]) -> str:
     name = task.get("name")
     if name:
-        return name
+        return str(name)
     action = task.get("action")
-    args = " ".join([u"{0}={1}".format(k, v) for (k, v) in action.items()
-                     if k not in ["__ansible_module__", "__ansible_arguments__"]] +
-                    action.get("__ansible_arguments__"))
+    if isinstance(action, str) or not isinstance(action, dict):
+        return str(action)
+    args = " ".join(
+        [
+            "{0}={1}".format(k, v)
+            for (k, v) in action.items()
+            if k
+            not in [
+                "__ansible_module__",
+                "__ansible_arguments__",
+                "__line__",
+                "__file__",
+            ]
+        ]
+    )
+    for item in action.get("__ansible_arguments__", []):
+        args += f" {item}"
     return u"{0} {1}".format(action["__ansible_module__"], args)
 
 
-def extract_from_list(blocks, candidates):
+def extract_from_list(blocks, candidates: List[str]) -> List[Any]:
     results = list()
     for block in blocks:
         for candidate in candidates:
@@ -562,12 +645,13 @@ def extract_from_list(blocks, candidates):
                     results.extend(add_action_type(block[candidate], candidate))
                 elif block[candidate] is not None:
                     raise RuntimeError(
-                        "Key '%s' defined, but bad value: '%s'" %
-                        (candidate, str(block[candidate])))
+                        "Key '%s' defined, but bad value: '%s'"
+                        % (candidate, str(block[candidate]))
+                    )
     return results
 
 
-def add_action_type(actions, action_type):
+def add_action_type(actions, action_type: str) -> List[Any]:
     results = list()
     for action in actions:
         # ignore empty task
@@ -578,25 +662,33 @@ def add_action_type(actions, action_type):
     return results
 
 
-def get_action_tasks(yaml, file):
+def get_action_tasks(yaml, file: Lintable) -> List[Any]:
     tasks = list()
-    if file['type'] in ['tasks', 'handlers']:
-        tasks = add_action_type(yaml, file['type'])
+    if file.kind in ['tasks', 'handlers']:
+        tasks = add_action_type(yaml, file.kind)
     else:
-        tasks.extend(extract_from_list(yaml, ['tasks', 'handlers', 'pre_tasks', 'post_tasks']))
+        tasks.extend(
+            extract_from_list(yaml, ['tasks', 'handlers', 'pre_tasks', 'post_tasks'])
+        )
 
     # Add sub-elements of block/rescue/always to tasks list
     tasks.extend(extract_from_list(tasks, ['block', 'rescue', 'always']))
     # Remove block/rescue/always elements from tasks list
     block_rescue_always = ('block', 'rescue', 'always')
-    tasks[:] = [task for task in tasks if all(k not in task for k in block_rescue_always)]
+    tasks[:] = [
+        task for task in tasks if all(k not in task for k in block_rescue_always)
+    ]
 
-    return [task for task in tasks if
-            set(['include', 'include_tasks',
-                'import_playbook', 'import_tasks']).isdisjoint(task.keys())]
+    return [
+        task
+        for task in tasks
+        if set(
+            ['include', 'include_tasks', 'import_playbook', 'import_tasks']
+        ).isdisjoint(task.keys())
+    ]
 
 
-def get_normalized_tasks(yaml, file):
+def get_normalized_tasks(yaml, file: Lintable) -> List[Dict[str, Any]]:
     tasks = get_action_tasks(yaml, file)
     res = []
     for task in tasks:
@@ -606,17 +698,18 @@ def get_normalized_tasks(yaml, file):
         if 'skip_ansible_lint' in (task.get('tags') or []):
             # No need to normalize_task is we are skipping it.
             continue
-        res.append(normalize_task(task, file['path']))
+        res.append(normalize_task(task, str(file.path)))
 
     return res
 
 
 @lru_cache(maxsize=128)
-def parse_yaml_linenumbers(data, filename):
+def parse_yaml_linenumbers(lintable: Lintable) -> AnsibleBaseYAMLObject:
     """Parse yaml as ansible.utils.parse_yaml but with linenumbers.
 
     The line numbers are stored in each node's LINE_NUMBER_KEY key.
     """
+
     def compose_node(parent, index):
         # the line number where the previous token has ended (plus empty lines)
         line = loader.line
@@ -630,23 +723,23 @@ def parse_yaml_linenumbers(data, filename):
             mapping[LINE_NUMBER_KEY] = node.__line__
         else:
             mapping[LINE_NUMBER_KEY] = mapping._line_number
-        mapping[FILENAME_KEY] = filename
+        mapping[FILENAME_KEY] = lintable.path
         return mapping
 
     try:
         kwargs = {}
         if 'vault_password' in inspect.getfullargspec(AnsibleLoader.__init__).args:
             kwargs['vault_password'] = DEFAULT_VAULT_PASSWORD
-        loader = AnsibleLoader(data, **kwargs)
+        loader = AnsibleLoader(lintable.content, **kwargs)
         loader.compose_node = compose_node
         loader.construct_mapping = construct_mapping
         data = loader.get_single_data()
     except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
-        raise SystemExit("Failed to parse YAML in %s: %s" % (filename, str(e)))
+        raise SystemExit("Failed to parse YAML in %s: %s" % (lintable.path, str(e)))
     return data
 
 
-def get_first_cmd_arg(task):
+def get_first_cmd_arg(task: Dict[str, Any]) -> Any:
     try:
         if 'cmd' in task['action']:
             first_cmd_arg = task['action']['cmd'].split()[0]
@@ -672,7 +765,7 @@ def is_playbook(filename: str) -> bool:
         "import_playbook",
         "post_tasks",
         "pre_tasks",
-        "roles"
+        "roles",
         "tasks",
     }
 
@@ -684,153 +777,94 @@ def is_playbook(filename: str) -> bool:
         f = parse_yaml_from_file(filename)
     except Exception as e:
         _logger.warning(
-            "Failed to load %s with %s, assuming is not a playbook.",
-            filename, e)
+            "Failed to load %s with %s, assuming is not a playbook.", filename, e
+        )
     else:
         if (
-            isinstance(f, AnsibleSequence) and
-            hasattr(f, 'keys') and
-            playbooks_keys.intersection(next(iter(f), {}).keys())
+            isinstance(f, AnsibleSequence)
+            and hasattr(next(iter(f), {}), 'keys')
+            and playbooks_keys.intersection(next(iter(f), {}).keys())
         ):
             return True
     return False
 
 
-def get_yaml_files(options: Namespace) -> dict:
-    """Find all yaml files."""
-    # git is preferred as it also considers .gitignore
-    git_command = ['git', 'ls-files', '*.yaml', '*.yml']
-    _logger.info("Discovering files to lint: %s", ' '.join(git_command))
+# pylint: disable=too-many-statements
+def get_lintables(
+    options: Namespace = Namespace(), args: Optional[List[str]] = None
+) -> List[Lintable]:
+    """Detect files and directories that are lintable."""
+    lintables: List[Lintable] = []
 
-    out = None
+    # passing args bypass auto-detection mode
+    if args:
+        for arg in args:
+            lintable = Lintable(arg)
+            if lintable.kind in ("yaml", None):
+                _logger.warning(
+                    "Overriding detected file kind '%s' with 'playbook' "
+                    "for given positional argument: %s",
+                    lintable.kind,
+                    arg,
+                )
+                lintable = Lintable(arg, kind="playbook")
+            lintables.append(lintable)
+    else:
 
-    try:
-        out = subprocess.check_output(
-            git_command,
-            stderr=subprocess.STDOUT,
-            universal_newlines=True
-        ).split()
-    except subprocess.CalledProcessError as exc:
-        _logger.warning(
-            "Failed to discover yaml files to lint using git: %s",
-            exc.output.rstrip('\n')
-        )
-    except FileNotFoundError as exc:
-        if options.verbosity:
-            _logger.warning(
-                "Failed to locate command: %s", exc
-            )
+        for filename in get_yaml_files(options):
 
-    if out is None:
-        out = [
-            os.path.join(root, name)
-            for root, dirs, files in os.walk('.')
-            for name in files
-            if name.endswith('.yaml') or name.endswith('.yml')
-        ]
-
-    return OrderedDict.fromkeys(sorted(out))
-
-
-# FIXME: drop noqa once this function is made simpler
-# Ref: https://github.com/ansible/ansible-lint/issues/744
-def get_playbooks_and_roles(options=None) -> List[str]:  # noqa: C901
-    """Find roles and playbooks."""
-    if options is None:
-        options = {}
-
-    files = get_yaml_files(options)
-
-    playbooks = []
-    role_dirs = []
-    role_internals = {
-        'defaults',
-        'files',
-        'handlers',
-        'meta',
-        'tasks',
-        'templates',
-        'vars',
-    }
-
-    # detect role in repository root:
-    if 'tasks/main.yml' in files or 'tasks/main.yaml' in files:
-        role_dirs.append('.')
-
-    for p in map(Path, files):
-
-        try:
-            for file_path in options.exclude_paths:
-                if str(p.resolve()).startswith(str(file_path)):
-                    raise FileNotFoundError(
-                        f'File {file_path} matched exclusion entry: {p}')
-        except FileNotFoundError as e:
-            _logger.debug('Ignored %s due to: %s', p, e)
-            continue
-
-        if (next((i for i in p.parts if i.endswith('playbooks')), None) or
-                'playbook' in p.parts[-1]):
-            playbooks.append(normpath(p))
-            continue
-
-        # ignore if any folder ends with _vars
-        if next((i for i in p.parts if i.endswith('_vars')), None):
-            continue
-        elif 'roles' in p.parts or '.' in role_dirs:
-            if 'tasks' in p.parts and p.parts[-1] in ['main.yaml', 'main.yml']:
-                role_dirs.append(str(p.parents[1]))
+            p = Path(filename)
+            # skip exclusions
+            try:
+                for file_path in options.exclude_paths:
+                    if str(p.resolve()).startswith(str(file_path)):
+                        raise FileNotFoundError(
+                            f'File {file_path} matched exclusion entry: {p}'
+                        )
+            except FileNotFoundError as e:
+                _logger.debug('Ignored %s due to: %s', p, e)
                 continue
-            elif role_internals.intersection(p.parts):
-                continue
-            elif 'tests' in p.parts:
-                playbooks.append(normpath(p))
-        if 'molecule' in p.parts:
-            if p.parts[-1] != 'molecule.yml':
-                playbooks.append(normpath(p))
-            continue
-        # hidden files are clearly not playbooks, likely config files.
-        if p.parts[-1].startswith('.'):
-            continue
 
-        if is_playbook(str(p)):
-            playbooks.append(normpath(p))
-            continue
+            lintables.append(Lintable(p))
 
-        _logger.info('Unknown file type: %s', normpath(p))
+        # stage 2: guess roles from current lintables, as there is no unique
+        # file that must be present in any kind of role.
+        _extend_with_roles(lintables)
 
-    _logger.info('Found roles: %s', ' '.join(role_dirs))
-    _logger.info('Found playbooks: %s', ' '.join(playbooks))
-
-    return role_dirs + playbooks
+    return lintables
 
 
-def expand_path_vars(path: str) -> str:
-    """Expand the environment or ~ variables in a path string."""
-    # It may be possible for function to be called with a Path object
-    path = str(path).strip()
-    path = os.path.expanduser(path)
-    path = os.path.expandvars(path)
-    return path
+def _extend_with_roles(lintables: List[Lintable]) -> None:
+    """Detect roles among lintables and adds them to the list."""
+    for lintable in lintables:
+        parts = lintable.path.parent.parts
+        if 'roles' in parts:
+            role = lintable.path
+            while role.parent.name != "roles" and role.name:
+                role = role.parent
+            if role.exists:
+                lintable = Lintable(role, kind="role")
+                if lintable not in lintables:
+                    _logger.debug("Added role: %s", lintable)
+                    lintables.append(lintable)
 
 
-def expand_paths_vars(paths: List[str]) -> List[str]:
-    """Expand the environment or ~ variables in a list."""
-    paths = [expand_path_vars(p) for p in paths]
-    return paths
+def convert_to_boolean(value: Any) -> bool:
+    """Use Ansible to convert something to a boolean."""
+    return bool(boolean(value))
 
 
-def get_rules_dirs(rulesdir: List[str], use_default: bool) -> List[str]:
-    """Return a list of rules dirs."""
-    default_ruledirs = [DEFAULT_RULESDIR]
-    default_custom_rulesdir = os.environ.get(
-        CUSTOM_RULESDIR_ENVVAR, os.path.join(DEFAULT_RULESDIR, "custom")
-    )
-    custom_ruledirs = sorted(
-        str(rdir.resolve())
-        for rdir in Path(default_custom_rulesdir).iterdir()
-        if rdir.is_dir() and (rdir / "__init__.py").exists()
-    )
-    if use_default:
-        return rulesdir + custom_ruledirs + default_ruledirs
-
-    return rulesdir or custom_ruledirs + default_ruledirs
+def nested_items(
+    data: Union[Dict[Any, Any], List[Any]]
+) -> Generator[Tuple[Any, Any], None, None]:
+    """Iterate a nested data structure."""
+    if isinstance(data, dict):
+        for k, v in data.items():
+            yield k, v
+            for k, v in nested_items(v):
+                yield k, v
+    if isinstance(data, list):
+        for item in data:
+            yield "list-item", item
+            for k, v in nested_items(item):
+                yield k, v
